# --------------------------------------------------------------------------
# Copyright 1992-1993 by Forschungszentrum Informatik (FZI)
#
# You can use and distribute this software under the terms of the license
# version 1 you should have received along with this software.
# If not or if you want additional information, write to
# Forschungszentrum Informatik, "STONE", Haid-und-Neu-Strasse 10-14,
# D-76131 Karlsruhe, Germany.
# --------------------------------------------------------------------------

###################################################################################
# Module:	ObjectKernel
# Author:	Rainer Neumann
# Date  :	July 22, 1993
# State :	finished
###################################################################################

if { [info exists _ok_ObjectKernel_] == 1 } {
set load_message "module ObjectKernel is already present"
} else {
set _ok_ObjectKernel_ included
set load_message "module ObjectKernel loading"

  # to provide multiple inheritence the two procedure _ok_call and _ok_getAllBaseclasses
  # must be changed (May 25, 1993)
  # also the ok_method procedure must be changed! (July 22, 1993)

###################
# Basic functions #
###################

proc Error { sender { error "" } } {
  if { $error == "" } { set error "Unspecified error" }
  error "Errormessage from $sender : $error"
}

set _ok_ObjectCount 0
set _ok_Classlist {}

proc NULL { method { args {} } } {
  Error {NULL} {Trying to access NULL-Object}
}

proc ok_call { object class method { parameter {} } } {
  upvar #0 _ok_${class}_methods MethTab
  set result ""
  if { [catch "set implClass $MethTab($method)"] } {
    Error ok_call "Calling undefined method $method out of class $class"
  } else {
    set result [eval "$implClass::$method $object $parameter"]
  }
  return $result
}

proc ok_newObject { } {
  global _ok_ObjectCount
  set objID "ObjRef_$_ok_ObjectCount"
  incr _ok_ObjectCount
  return $objID
}

proc ok_getAllBaseclasses { class } {
  global _ok_Classinfo _ok_Classlist
  if { [lsearch $_ok_Classlist $class] > -1 } {
    set result {}
    if { $class != "" } {
      set bc [lindex $_ok_Classinfo($class) 0]
    }
    while { $bc != "" } {
      set result [concat $result $bc]
      set bc [lindex $_ok_Classinfo($bc) 0]
    }
  } else {
    Error ObjectKernel "No class with name $class defined!"
  }
  return $result
}

proc ok_getClassAttributes { class } {
  global _ok_Classinfo
  return [lindex $_ok_Classinfo($class) 1]
}

proc ok_getClassAttributeDefaults { class } {
  global _ok_Classinfo
  return [lindex $_ok_Classinfo($class) 2]
}

proc ok_getAllAttributes { class } {
  set result ""
  if { $class != "" } {
    set rl1 {}
    set rl2 {}
    foreach bc [concat $class [ok_getAllBaseclasses $class]] {
      set rl1 [concat [ok_getClassAttributes $bc] $rl1]
      set rl2 [concat [ok_getClassAttributeDefaults $bc] $rl2]
    }
  }
  return [list $rl1 $rl2]
}

##################################
# Function to create a new class #
##################################

proc ok_class { classname baseclasses attributes defaults } {
  global _ok_Classinfo _ok_Classlist _ok_DerivedClasses
  set _ok_DerivedClasses($classname) {}
  upvar #0 _ok_${classname}_methods MethTab
  if { ($classname != "Object") && ([lsearch $_ok_Classlist $classname] != -1) } {
    Error ok_class "Class $classname is already defined"
  }
  foreach bc $baseclasses {
    if { [lsearch $_ok_Classlist $bc] == -1 } {
      Error ok_class "Baseclass $bc of class $classname is undefined"
    } else {
      set _ok_DerivedClasses($bc) [concat $_ok_DerivedClasses($bc) $classname]
      upvar #0 _ok_${bc}_methods MethTab_$bc
      foreach meth [array names "MethTab_$bc"] {
	set hs "($meth)"
        set hs "MethTab_$bc$hs"
	set MethTab($meth) [set $hs]
      }
    }
  }
  if { [lsearch $_ok_Classlist $classname] == -1 } {
    lappend _ok_Classlist $classname
  }
  set _ok_Classinfo($classname) [list $baseclasses $attributes $defaults]
  set ai ""
  set i 0
  foreach attr $attributes {
    # define access-functions get_attr and set_attr
    set p "ok_method $classname get_$attr {} {\n  return \$attr($attr)\n}"
    eval $p
    set p "ok_method $classname set_$attr { { [concat _$attr [list [lindex $defaults $i]]] } } {\n  set attr($attr) \$_$attr \n  return \$self\n}"
    eval $p
    set ai "$ai  set attr($attr) [list [lindex $defaults $i]]\n"
    incr i
  }
  set p ""
  foreach bc $baseclasses { set p "$p\n  $bc::destroy \$self" }
  ok_destructor $classname {} $p
  return "Created class $classname"
}

proc ok_method { classname methodname arglist body } {
  global _ok_Classlist _ok_Classinfo _ok_DerivedClasses
  upvar #0 _ok_${classname}_methods MethodTable
  if { $methodname != "create" } {
    set arglist [concat { self } $arglist]
    set newbody "  upvar #0 \$self attr\n$body"
  } else {
    set newbody $body
  }
  eval "proc $classname::$methodname { $arglist } { $newbody }"
  # Now write the method information into method table
  if { $methodname != "create" } {
    set MethodTable($methodname) $classname
    set dc $_ok_DerivedClasses($classname)
    set visited {}
    set bcs [ok_getAllBaseclasses $classname]
    while { $dc != {} } {
      set newdc {}
      foreach cl $dc {
        upvar #0 "${cl}_methods" meth_$cl
        set hs "($methodname)"
        set hs "meth_$cl$hs"
        if { [catch "set implClass [set $hs]"] } {
	  # Method was not known in that class
          set "{${cl}_methods}($methodname)" $classname
	  set newdc [concat $newdc $DerivedClasses($cl)]
        } else {
          if { ([lsearch $visited $implClass] == -1) &&
	       ([lsearch $bcs $implClass] > -1) } {
	    set [set hs] $classname
            set newdc [concat $newdc $_ok_DerivedClasses($cl)]
          }
        }
        lappend visited $cl
      }
      set dc $newdc
    }
  }
}

proc ok_abstractmethod { classname methodname  arglist } {
  ok_method $classname $methodname $arglist "\n  Error $classname::$methodname {Calling abstract method}\n"
}

proc ok_constructor { classname arglist body } {
  ok_method $classname create $arglist $body
}

proc ok_destructor { classname arglist body } {
  ok_method $classname destroy $arglist $body
}


################
# Class Object #
################

ok_class Object "" { Class } { Object }

ok_constructor Object { classname } {
  if { $classname == "" } { set classname Object }
  set objID [ok_newObject]
  set newObj [list $classname {}]
  upvar #0 $objID attr
  set at [ok_getAllAttributes $classname]
  set al [lindex $at 0]
  set dl [lindex $at 1]
  set i 0
  foreach a $al {
    set attr($a) [lindex $dl $i]
    incr i
  }
  set attr(Class) $classname
  eval "proc $objID { method args } { ok_call $objID $attr(Class) \$method \$args }"
  return $objID
}

ok_destructor Object {} {
  foreach a [lindex [ok_getAllAttributes $attr(Class)] 0] {
    set attr($a) ""
    unset attr($a)
  }
  set attr(Class) Removed
}

ok_method Object isRemoved {} {
  if { $attr(Class) == "Removed" } { set r 1 } else { set r 0 }
  return $r
}

ok_method Object type {} {
  return $attr(Class)
}

ok_method Object isA { classname } {
  global Classinfo
  if { [lsearch [ok_getAllBaseclasses $attr(Class)] $classname] > -1 } {
    set result 1
  } else {
    set result 0
  }
  return $result
}

ok_method Object copy {} {
  set type [Object::type $self]
  set newObj [Object::create $type]
  upvar #0 $newObj newAttr
  foreach a [lindex [ok_getAllAttributes $type] 0] {
    set newAttr($a) $attr($a)
  }
  return $newObj
}

ok_method Object convert { newClass } {
  global _ok_Classlist
  if { [lsearch $_ok_Classlist $newClass] > -1 } {
    set at1 [ok_getAllAttributes [Object::type $self]]
    set al1 [lindex $at1 0]
    set at2 [ok_getAllAttributes $newClass]
    set al2 [lindex $at2 0]
    set dl2 [lindex $at2 1]
    set il [intersect3 $al1 $al2]
    foreach a [lindex $il 0] {
      unset attr($a)
    }
    foreach a [lindex $il 2] {
      set attr($a) [lindex $dl2 [lsearch $al2 $a]]
    }
    set attr(Class) $newClass
  } else {
    Error Object::convert \
		"Cannot convert class [Object::type $self] to non-class $newClass!"
  }
}


#######################
# Pseudoclass Removed #
#######################

ok_class Removed Object {} {}


# if the used interpreter doesn't support extended tcl, the following function
# is automatically declared

if { [info command intersect3] == "" } {

  proc intersect3 { list1 list2 } {
    set rl1 {}
    set rl2 {}
    set rl3 {}
    foreach x $list1 {
      if { [lsearch $list2 $x] == -1 } {
	if { [lsearch $rl1 $x] == -1 } {
          lappend rl1 $x
        }
      } else {
        if { [lsearch $rl2 $x] == -1 } {
          lappend rl2 $x
        }
      }
    }
    foreach y $list2 {
      if { [lsearch $list1 $y] == -1 } {
        if { [lsearch $rl3 $y] == -1 } {
          lappend rl3 $y
        }
      }
    }
    return [list $rl1 $rl2 $rl3]
  }

}


set load_message "module ObjectKernel loaded"
}
